/**************************************
 * Direct video memory display driver *
 * Saku Airila 1996                   *
 **************************************/

#include <linuxmt/types.h>
#include <linuxmt/config.h>
#include <linuxmt/errno.h>
#include <linuxmt/fcntl.h>
#include <linuxmt/fs.h>
#include <linuxmt/major.h>

#ifdef CONFIG_CONSOLE_DIRECT

#define CONSOLE_NAME "console"

#define WAIT_MV1_PARM 1
#define WAIT_MV2_PARM 2
#define WAIT_P_PARM 3
#define WAIT_Q_PARM 4

#define MAX_ATTR 3

#define A_DEFAULT 0x07
#define A_BOLD 0x08
#define A_BLINK 0x80

static struct
{
   int w, h;
   int cx, cy;
   int CCBase;
   int InCmd;
   int WaitParm, PTmp;
   unsigned VideoSeg;
   unsigned CursorOfs;
   unsigned char Attr;
} Scr;

static unsigned AttrArry[] = {
   A_DEFAULT,
   A_BOLD,
   A_BLINK
   };

static void ScrollUp();
static void ScrollDown();
static void PositionCursor();
static void WriteChar();
static void CommandChar();
static void ClearRange();
static int GetParm();
static void WaitParameter();

extern int peekb();
extern int peekw();
extern void pokeb();
extern void pokew();
extern void outb();

extern int AddQueue(); /* From new_key.c */
extern int GetQueue(); 

void Console_gotoxy();  /* Does this need be exported? */

/* This will probably go away, but currently printk() uses it */
void con_charout( Ch )
char Ch;
{
   WriteChar( Ch );
   PositionCursor(Scr.cx,Scr.cy);
}

void WriteChar( Ch )
char Ch;
{
   if( Scr.InCmd )
   {
      CommandChar( Ch );
      return;
   }
   if( Ch == 27 )
   {
      Scr.InCmd = 1;
      return;
   }
   if( Ch == '\b' )
   {
      if( Scr.cx > 0 )
      {
         Scr.cx--;
         WriteChar( ' ' );
         Scr.cx--;
         PositionCursor(Scr.cx,Scr.cy);
      }
      return;
   }
   if( Ch == '\n' )
      Scr.cy++;
   else if( Ch == '\r' )
      Scr.cx = 0;
   else
   {
      Scr.CursorOfs = ( Scr.cx << 1 ) + ( ( Scr.w * Scr.cy ) << 1 );
      pokeb( Scr.VideoSeg, Scr.CursorOfs, Ch );
      pokeb( Scr.VideoSeg, Scr.CursorOfs + 1, Scr.Attr );
      Scr.cx++;
   }
   if( Scr.cx >= Scr.w - 1 )
      Scr.cx = Scr.w - 1;
   if( Scr.cy == Scr.h )
   {
      ScrollUp( 1, Scr.h );
      Scr.cy--;
   }
}

void ScrollUp( st, en )
int st, en;
{
unsigned rdofs, wrofs, SegAddr;
int cnt;
   if( st > en || st < 1 )
      return;
   rdofs = ( Scr.w << 1 ) * st;
   wrofs = rdofs - ( Scr.w << 1 );
   cnt = ( Scr.w * ( en - st ) );
   SegAddr = Scr.VideoSeg;
   for( ; cnt--; wrofs += 2, rdofs += 2 )
      pokew( SegAddr, wrofs, peekw( SegAddr, rdofs ) );
   ClearRange( 0, Scr.h - 1, Scr.w, Scr.h - 1 );
}

void ScrollDown( st, en )
int st, en;
{
unsigned rdofs, wrofs;
int cnt;
   if( st > en || en > Scr.h )
      return;
   cnt = ( Scr.w * ( en - st ) );
   rdofs = ( Scr.w << 1 ) * st + ( cnt << 1 );
   wrofs = rdofs + ( Scr.w << 1 );
   for( ; cnt-- > -1; wrofs -= 2, rdofs -= 2 )
      pokew( Scr.VideoSeg, wrofs, peekw( Scr.VideoSeg, rdofs ) );
   ClearRange( 0, st, Scr.w, st );
}

void PositionCursor( x, y )
int x, y;
{
int Pos;
   Pos = x + Scr.w * y;
   outb( 14, Scr.CCBase );
   outb( ( unsigned char )( ( Pos >> 8 ) & 0xFF ), Scr.CCBase + 1 );
   outb( 15, Scr.CCBase );
   outb( ( unsigned char )( Pos & 0xFF ), Scr.CCBase + 1 );
}

void ClearRange( x, y, xx, yy )
int x, y, xx, yy;
{
unsigned st, en, ClrW;
   st = ( x << 1 ) + y * ( Scr.w << 1 );
   en = ( xx << 1 ) + yy * ( Scr.w << 1 );
   ClrW = A_DEFAULT << 8;
   for( ; st < en; st += 2 )
      pokew( Scr.VideoSeg, st, ClrW );
}

void CommandChar( Ch )
char Ch;
{
   if( Scr.WaitParm )
   {
      WaitParameter( Ch );
      return;
   }
   switch( Ch )
   {
      case 'H' : /* home */
         Scr.cx = Scr.cy = 0;
         break;
      case 'A' : /* up */
         Scr.cy = Scr.cy ? Scr.cy - 1 : 0;
         break;
      case 'B' : /* down */
         Scr.cy = Scr.cy < Scr.h - 1 ? Scr.cy + 1 : Scr.h - 1;
         break;
      case 'C' : /* right */
         Scr.cx = Scr.cx < Scr.w - 1 ? Scr.cx + 1 : Scr.w - 1;
         break;
      case 'D' : /* left */
         Scr.cx = Scr.cx ? Scr.cx - 1 : 0;
         break;
      case 'K' : /* clear to eol */
         ClearRange( Scr.cx, Scr.cy, Scr.w, Scr.cy );
         break;
      case 'J' : /* clear to eoscreen */
         ClearRange( 0, 0, Scr.w, Scr.h );
         break;
      case 'I' : /* linefeed reverse */
         ScrollDown( 0, Scr.h - 1 );
         break;
      case 'Z' : /* identify */
         AddQueue( 27 );
         AddQueue( 'Z' );
         break;
      case 'P' : /* NOT VT52. insert/delete line (VT 61)*/
         Scr.WaitParm = WAIT_P_PARM;
         return;
      case 'Q' : /* NOT VT52. My own additions. Attributes... */
         Scr.WaitParm = WAIT_Q_PARM;
         return;
      case 'Y' : /* cursor move */
         Scr.WaitParm = WAIT_MV1_PARM;
         return;
   }
   Scr.InCmd = 0;
}

void WaitParameter( Ch )
char Ch;
{
   switch( Scr.WaitParm )
   {
      case WAIT_MV1_PARM :
         Scr.PTmp = Ch - ' ';
         Scr.WaitParm = WAIT_MV2_PARM;
         return;
      case WAIT_MV2_PARM :
         Console_gotoxy( Ch - ' ', Scr.PTmp );
         break;
      case WAIT_P_PARM :
         switch( Ch )
         {
            case 'f' : /* insert line at crsr */
               ScrollDown( Scr.cy, Scr.h - 1 );
               break;
            case 'd' : /* delete line at crsr */
               ScrollUp( Scr.cy + 1, Scr.h );
               ClearRange( 0, Scr.h, Scr.w, Scr.h );
               break;
         }
         break;
      case WAIT_Q_PARM :
         Scr.PTmp = Ch - ' ';
         if( Scr.PTmp > 0 && Scr.PTmp < MAX_ATTR )
            Scr.Attr |= AttrArry[ Scr.PTmp ];
         if( !Scr.PTmp )
            Scr.Attr = A_DEFAULT;
         break;
      default :
         break;
   }
   Scr.WaitParm = Scr.InCmd = 0;
}

void Console_gotoxy( x, y )
int x, y;
{
   if( x >= Scr.w )
      x = Scr.w - 1;
   if( y >= Scr.h )
      y = Scr.h - 1;
   if( x < 0 )
      x = 0;
   if( y < 0 )
      y = 0;
   Scr.cx = x;
   Scr.cy = y;
   PositionCursor( x, y );
}

int Console_lseek( inode, file, offset, origin )
struct inode *inode;
struct file *file;
off_t offset;
int origin;
{
return -ESPIPE;
}

int Console_write( inode, file, buf, count )
struct inode *inode;
struct file *file;
char *buf;
int count;
{
int cnt;
   for( cnt = 0; cnt < count; cnt++ )
      WriteChar( *buf++ );
   PositionCursor( Scr.cx, Scr.cy );
return cnt;
}

int Console_read( inode, file, buf, count )
struct inode *inode;
struct file *file;
char *buf;
int count;
{
int Ch1, cnt = 0;
   Ch1 = GetQueue();
   if( file->f_flags & O_NONBLOCK )
      if( Ch1 == -1 )
         return -EWOULDBLOCK;
      else
      {
         do
         {
            *buf++ = Ch1;
            Ch1 = GetQueue();
            cnt++;
         } while( Ch1 != -1 );
         return cnt;
      }
/* So there. Now, if there's nothing to read, and the file is not in
   non-blocking mode, how do we block the process? */
#if 0
   do
   {
      *buf++ = Ch1;
      Ch1 = GetQueue();
      cnt++;
   } while( Ch1 != -1 );
   return cnt;
#endif
return -EINVAL;
}

void Console_release( inode, file )
struct inode *inode;
struct file *file;
{
return;
}

int Console_open( inode, file )
struct inode *inode;
struct file *file;
{
int minor = MINOR(inode->i_rdev);
   if( minor )
      return -ENODEV;
return 0;
}

int Console_readdir()
{
return -ENOTDIR; /* Is this ok, or should it be -EBADF ? */
}

static struct file_operations ConOps =
{
   Console_lseek,
   Console_read,
   Console_write,
   Console_readdir, /* readdir */
   NULL,            /* Write select later */
   NULL,            /* Write ioctl later */
   Console_open,    /* Open the direct console */
   Console_release, /* Close the direct console */
   NULL,            /* Fsync */
   NULL,            /* No media */
   NULL             /* No media */
};

void init_console()
{
   Scr.w = peekb( 0x40, 0x4a );
   Scr.h = 25; /* Trust this. Cga does not support peeking at 0x40:0x84. */
   Scr.CCBase = peekw( 0x40, 0x63 );
   if( peekb( 0x40, 0x49 ) == 7 )
         Scr.VideoSeg = 0xB000;
   else
      Scr.VideoSeg = 0xB800;
   Scr.cx = peekb( 0x40, 0x50 );
   Scr.cy = peekb( 0x40, 0x51 );
   Scr.CursorOfs = 0;
   Scr.InCmd = 0;
   Scr.Attr = A_DEFAULT;
   if( register_chrdev(TTY_MAJOR, CONSOLE_NAME, &ConOps ) )
      printk( "Could not register console device.\n" );
   else
      printk(
         "Console: Direct %dx%d Mem@%x CrtC@%x emulating vt52\n",
         Scr.w,
         Scr.h,
         Scr.VideoSeg,
         Scr.CCBase );
}

#endif

